@@ -6,9 +6,6 @@ use crate::traits::{SCFunMF, SCFunRF, Sss};
 
 const DEFAULT_CONTEXT: () = ();
 
-// type BoxMF<Ctx, Req, Res> = Box<dyn Fn(&Ctx, usize, &Req) -> Res>;
-type BoxRF<Ctx, Res> = Box<dyn Fn(&Ctx, Res, Res) -> Res>;
-
 pub struct Manager<'a, Ctx> {
     context: &'a Ctx,
 }
@@ -31,8 +28,18 @@ pub struct ReduceManager<
     fun: RF,
 }
 
-type ReduceRManager<'a, Ctx, Req, ResOk, ResErr, MF> =
-    ReduceManager<'a, Ctx, Req, Result<ResOk, ResErr>, MF, BoxRF<Ctx, Result<ResOk, ResErr>>>;
+pub struct ReduceRManager<
+    'a,
+    Ctx,
+    Req,
+    ResOk: Sss,
+    ResErr: Sss,
+    MF: Fn(&Ctx, usize, &Req) -> Result<ResOk, ResErr>,
+    RF: Fn(&Ctx, ResOk, ResOk) -> Result<ResOk, ResErr>,
+> {
+    map_manager: MapManager<'a, Ctx, Req, Result<ResOk, ResErr>, MF>,
+    fun: RF,
+}
 
 pub struct FullManager<
     'a,
@@ -46,6 +53,19 @@ pub struct FullManager<
     default_value: Resp,
 }
 
+pub struct FullRManager<
+    'a,
+    Ctx,
+    Req,
+    ROk: Sss,
+    RErr: Sss,
+    MF: Fn(&Ctx, usize, &Req) -> Result<ROk, RErr>,
+    RF: Fn(&Ctx, ROk, ROk) -> Result<ROk, RErr>,
+> {
+    reduce_manager: ReduceRManager<'a, Ctx, Req, ROk, RErr, MF, RF>,
+    default_value: ROk,
+}
+
 pub fn manager() -> Manager<'static, ()> {
     Manager {
         context: &DEFAULT_CONTEXT,
@@ -81,26 +101,16 @@ impl<'a, Ctx, Req, Resp: Sss, MF: SCFunMF<Ctx, Req, Resp>> MapManager<'a, Ctx, R
     }
 }
 
-#[inline]
-fn get_ok<O, E>(res: Result<O, E>) -> O {
-    match res {
-        Ok(val) => val,
-        Err(_) => panic!("err in reduce function"),
-    }
-}
-
 impl<'a, Ctx, Req, ROk: Sss, RErr: Sss, MF: SCFunMF<Ctx, Req, Result<ROk, RErr>>>
     MapManager<'a, Ctx, Req, Result<ROk, RErr>, MF>
 {
     pub fn reduce_result<RF: 'static + Fn(&Ctx, ROk, ROk) -> Result<ROk, RErr>>(
         self,
         fun: RF,
-    ) -> ReduceRManager<'a, Ctx, Req, ROk, RErr, MF> {
-        let new_fun: BoxRF<Ctx, Result<ROk, RErr>> =
-            Box::new(move |ctx, a, b| fun(ctx, get_ok(a), get_ok(b)));
-        ReduceManager {
+    ) -> ReduceRManager<'a, Ctx, Req, ROk, RErr, MF, RF> {
+        ReduceRManager {
             map_manager: self,
-            fun: new_fun,
+            fun,
         }
     }
 }
@@ -131,15 +141,26 @@ impl<
         RF: 'static + Fn(&Ctx, Resp, Resp) -> Resp,
     > ReduceManager<'a, Ctx, Req, Resp, MF, RF>
 {
-    fn into_full(self) -> FullManager<'a, Ctx, Req, Resp, MF, RF> {
-        FullManager {
-            reduce_manager: self,
-            default_value: Resp::default(),
-        }
-    }
-
     pub fn run(self, chunks: &[Req]) -> Result<Resp, ThreadError> {
-        self.into_full().run(chunks)
+        self.default(Resp::default()).run(chunks)
+    }
+}
+
+impl<
+        'a,
+        Ctx,
+        Req,
+        ROk: Sss,
+        RErr: Sss,
+        MF: SCFunMF<Ctx, Req, Result<ROk, RErr>>,
+        RF: 'static + Fn(&Ctx, ROk, ROk) -> Result<ROk, RErr>,
+    > ReduceRManager<'a, Ctx, Req, ROk, RErr, MF, RF>
+{
+    fn default(self, default_value: ROk) -> FullRManager<'a, Ctx, Req, ROk, RErr, MF, RF> {
+        FullRManager {
+            reduce_manager: self,
+            default_value,
+        }
     }
 }
 
@@ -150,14 +171,34 @@ impl<
         ROk: Sss + Default,
         RErr: Sss,
         MF: SCFunMF<Ctx, Req, Result<ROk, RErr>>,
-        RF: 'static + Fn(&Ctx, Result<ROk, RErr>, Result<ROk, RErr>) -> Result<ROk, RErr>,
-    > ReduceManager<'a, Ctx, Req, Result<ROk, RErr>, MF, RF>
+        RF: 'static + Fn(&Ctx, ROk, ROk) -> Result<ROk, RErr>,
+    > ReduceRManager<'a, Ctx, Req, ROk, RErr, MF, RF>
 {
-    pub fn run_result(self, chunks: &[Req]) -> Result<ROk, MapReduceError<RErr>> {
-        self.default(Ok(ROk::default())).run_result(chunks)
+    pub fn run(self, chunks: &[Req]) -> Result<ROk, MapReduceError<RErr>> {
+        self.default(ROk::default()).run(chunks)
     }
 }
 
+fn make_workers<Ctx, Req, Resp: Sss, MF: Fn(&Ctx, usize, &Req) -> Resp>(
+    map_manager: &MapManager<Ctx, Req, Resp, MF>,
+    chunks: &[Req],
+) -> Vec<Worker<Resp>> {
+    let fun = CtxWrapper::new(&map_manager.fun);
+    let ctx = CtxWrapper::new(map_manager.manager.context);
+    let mut workers = Vec::new();
+    for (id, chunk) in chunks.iter().enumerate() {
+        let request = CtxWrapper::new(chunk);
+        let handler: thread::JoinHandle<Resp> =
+            thread::spawn(move || fun.get::<MF>()(ctx.get::<Ctx>(), id, request.get::<Req>()));
+        let worker = Worker {
+            id,
+            thread: Box::new(handler),
+        };
+        workers.push(worker);
+    }
+    workers
+}
+
 impl<
         'a,
         Ctx,
@@ -167,25 +208,8 @@ impl<
         RF: 'static + Fn(&Ctx, Resp, Resp) -> Resp,
     > FullManager<'a, Ctx, Req, Resp, MF, RF>
 {
-    fn make_workers(&self, chunks: &[Req]) -> Vec<Worker<Resp>> {
-        let fun = CtxWrapper::new(&self.reduce_manager.map_manager.fun);
-        let ctx = CtxWrapper::new(self.reduce_manager.map_manager.manager.context);
-        let mut workers = Vec::new();
-        for (id, chunk) in chunks.iter().enumerate() {
-            let request = CtxWrapper::new(chunk);
-            let handler: thread::JoinHandle<Resp> =
-                thread::spawn(move || fun.get::<MF>()(ctx.get::<Ctx>(), id, request.get::<Req>()));
-            let worker = Worker {
-                id,
-                thread: Box::new(handler),
-            };
-            workers.push(worker);
-        }
-        workers
-    }
-
     pub fn run(self, chunks: &[Req]) -> Result<Resp, ThreadError> {
-        let mut workers = self.make_workers(chunks);
+        let mut workers = make_workers(&self.reduce_manager.map_manager, chunks);
         let mut result = self.default_value;
         let ctx = self.reduce_manager.map_manager.manager.context;
         let mut failed = None;
@@ -219,44 +243,38 @@ impl<
         ROk: Sss,
         RErr: Sss,
         MF: 'static + Fn(&Ctx, usize, &Req) -> Result<ROk, RErr>,
-        RF: 'static + Fn(&Ctx, Result<ROk, RErr>, Result<ROk, RErr>) -> Result<ROk, RErr>,
-    > FullManager<'a, Ctx, Req, Result<ROk, RErr>, MF, RF>
+        RF: 'static + Fn(&Ctx, ROk, ROk) -> Result<ROk, RErr>,
+    > FullRManager<'a, Ctx, Req, ROk, RErr, MF, RF>
 {
-    pub fn run_result(self, chunks: &[Req]) -> Result<ROk, MapReduceError<RErr>> {
-        let mut workers = self.make_workers(chunks);
+    fn do_map_reduce(
+        self,
+        workers: &mut Vec<Worker<Result<ROk, RErr>>>,
+    ) -> Result<ROk, MapReduceError<RErr>> {
         let ctx = self.reduce_manager.map_manager.manager.context;
         let mut result = self.default_value;
-        let mut failed = None;
-        workers.reverse();
         for _ in 0..workers.len() {
             let worker = workers.pop().unwrap();
             let worker_res = match worker.thread.join() {
-                Ok(val) => val,
-                Err(err) => {
-                    failed = Some(err);
-                    break;
-                }
+                Ok(Ok(val)) => val,
+                Ok(Err(err)) => return Err(MapReduceError::Custom(err)),
+                Err(err) => return Err(MapReduceError::ThreadFailed(err)),
             };
-            if worker_res.is_ok() {
-                result = (self.reduce_manager.fun)(ctx, result, worker_res);
-                if result.is_err() {
-                    break;
-                }
-            } else {
-                result = worker_res;
-                // stop reducing results
-                break;
+            match (self.reduce_manager.fun)(ctx, result, worker_res) {
+                Ok(val) => result = val,
+                Err(err) => return Err(MapReduceError::Custom(err)),
             }
         }
-        // skip rest results
-        for worker in workers {
-            let _ = worker.thread.join();
-        }
-        match (failed, result) {
-            (Some(err), _) => Err(MapReduceError::ThreadFailed(err)),
-            (_, Err(err)) => Err(MapReduceError::Custom(err)),
-            (_, Ok(val)) => Ok(val),
+        Ok(result)
+    }
+
+    pub fn run(self, chunks: &[Req]) -> Result<ROk, MapReduceError<RErr>> {
+        let mut workers = make_workers(&self.reduce_manager.map_manager, chunks);
+        workers.reverse();
+        let result = self.do_map_reduce(&mut workers);
+        for rest_worker in workers {
+            let _ = rest_worker.thread.join();
         }
+        result
     }
 }
 
@@ -321,7 +339,7 @@ mod tests {
         let res = manager()
             .map(|_, _, s: &&str| u8::from_str(s))
             .reduce_result(|_, a: u8, b: u8| Ok(a + b))
-            .run_result(&["1", "2", "3"]);
+            .run(&["1", "2", "3"]);
         match res {
             Ok(6) => (),
             _ => panic!(),
@@ -333,7 +351,7 @@ mod tests {
         let res = manager()
             .map(|_, _, s: &&str| u8::from_str(s))
             .reduce_result(|_, a: u8, b: u8| Ok(a + b))
-            .run_result(&["1", "arr", "3"]);
+            .run(&["1", "arr", "3"]);
         let err = match res {
             Err(MapReduceError::Custom(err)) => err.to_string(),
             _ => panic!(),
@@ -346,7 +364,7 @@ mod tests {
         let res = manager()
             .map(|_, _, s: &u8| Ok(*s))
             .reduce_result(|_, _: u8, _: u8| Err(()))
-            .run_result(&[1, 2, 3]);
+            .run(&[1, 2, 3]);
         match res {
             Err(MapReduceError::Custom(())) => (),
             _ => panic!(),
@@ -365,7 +383,7 @@ mod tests {
                 }
             })
             .reduce_result(|_, a, b| Ok(a + b))
-            .run_result(&[1, 1, 2, 3, 3]);
+            .run(&[1, 1, 2, 3, 3]);
         match res {
             Err(MapReduceError::Custom(())) => (),
             _ => panic!(),
@@ -404,6 +422,25 @@ mod tests {
         }
     }
 
+    fn func_res(context: &str, data: &[StructWrapper]) -> Result<Vec<String>, String> {
+        manager()
+            .context(&context)
+            .map(|ctx, _, val: &&[StructWrapper]| {
+                let mut sum = 0;
+                for s in *val {
+                    sum += s.data;
+                }
+                Ok(vec![format!("{}: {}", ctx, sum)])
+            })
+            .reduce_result(|_, mut a, mut b| {
+                a.append(&mut b);
+                let res: Result<Vec<String>, String> = Ok(a);
+                res
+            })
+            .run(&[&data[0..2], &data[2..4]])
+            .map_err(|_| "err".to_string())
+    }
+
     #[test]
     fn test_map_per_chunks() {
         let wrapped = vec![
@@ -415,4 +452,21 @@ mod tests {
         assert_eq!(func("hello", &wrapped), 10);
         assert_eq!(wrapped[0].data, 1);
     }
+
+    #[test]
+    fn test_map_result_per_chunks() {
+        let ctx = "hi";
+        let wrapped = vec![
+            StructWrapper { data: 1 },
+            StructWrapper { data: 2 },
+            StructWrapper { data: 3 },
+            StructWrapper { data: 4 },
+        ];
+        assert_eq!(
+            func_res(ctx, &wrapped),
+            Ok(vec!["hi: 3".to_string(), "hi: 7".to_string()])
+        );
+        assert_eq!(wrapped[0].data, 1);
+        assert_eq!(ctx, "hi");
+    }
 }
